import json
import pathlib

__here__ = pathlib.Path(__file__).parent

with open(__here__ / "HandshakeRequest.avsc", "r") as f:
    handshake_request = json.load(f)


with open(__here__ / "HandshakeResponse.avsc", "r") as f:
    handshake_response = json.load(f)



from collections.abc import Mapping


class DotDict(Mapping):
    """A simple mapping allowing dot notation."""

    def __init__(self, **kwargs):
        self._dict = kwargs
        for k, v in self._dict.items():
            setattr(self, k, v)

    def __getitem__(self, name):
        return self._dict[name]

    def __setitem__(self, k, v):
        self._dict[k] = v
        setattr(self, k, v)

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)



__all__ = ["Client", "YaqDaemonException"]


import functools
import inspect
import json
import struct
from threading import Lock
import types

from ._socket import Socket
from ._dotdict import DotDict
from ._properties import Property


class YaqDaemonException(Exception):
    pass


def reconnect(fun):
    """
    If the socket link is broken, try creating a new link and run the method again.
    """

    @functools.wraps(fun)
    def inner(self, *args, **kwargs):
        try:
            return fun(self, *args, **kwargs)
        except (ConnectionError, struct.error):
            self._socket = Socket(self._host, self._port)
            self.handshake()
            return fun(self, *args, **kwargs)

    return inner


class Client:
    def __init__(self, port, host="127.0.0.1"):
        self._host = host
        self._port = port
        self._socket = Socket(self._host, self._port)
        self._id_counter = 0
        self._connection_callbacks = []
        self._mutex = Lock()
        self._cached_id = dict()
        self.handshake()
        self.id()  # populates cached_id

    def __repr__(self):
        name = self._cached_id.get("name", "")
        protocol = self._protocol["protocol"]
        host = self._host
        port = self._port
        return f"<yaqc.Client to {host}:{port} ({protocol}:{name})>"

    def handshake(self):
        def fun(comm, sig):
            def inner(self, *args, **kwargs):
                ba = sig.bind_partial(self, *args, **kwargs)
                ba.apply_defaults()
                return self.send(comm, *ba.args[1:], **ba.kwargs)

            inner.__signature__ = sig
            inner.__name__ = comm
            return inner

        with self._mutex:
            self._protocol = json.loads(self._socket.handshake())
            self._named_types = {t["name"]: t for t in self._protocol.get("types", [])}
            self._socket._named_types = self._named_types
            self.traits = self._protocol["traits"]
            for name, props in self._protocol.get("messages", {}).items():
                if hasattr(self, name):
                    continue

                params = [inspect.Parameter("self", inspect.Parameter.POSITIONAL_ONLY)]
                for param in props.get("request", []):
                    params.append(
                        inspect.Parameter(
                            param["name"],
                            inspect.Parameter.POSITIONAL_OR_KEYWORD,
                            default=param.get("default", inspect.Parameter.empty),
                        )
                    )
                sig = inspect.Signature(params)
                method = fun(name, sig)
                method.__doc__ = props.get("doc")
                setattr(self, name, types.MethodType(method, self))

            # Must be done after above message setattrs
            self.properties = DotDict()
            for k, v in self._protocol.get("properties", {}).items():
                self.properties[k] = Property(self, v)
            for cb in self._connection_callbacks:
                cb()

    @reconnect
    def send(self, method, *args, **kwargs):
        with self._mutex:
            self._id_counter += 1
            return self._socket.message(
                method, self._protocol["messages"][method], *args, **kwargs
            )

    def register_connection_callback(self, fun):
        self._connection_callbacks.append(fun)

    def clear_connection_callbacks(self):
        self._connection_callbacks = []

    def id(self):
        self._cached_id = self.send("id")
        return self._cached_id



"""Define version."""


import pathlib
import subprocess


here = pathlib.Path(__file__).resolve().parent


__all__ = ["__version__", "__branch__"]


# read from VERSION file
with open(str(here / "VERSION")) as f:
    __version__ = f.read().strip()

try:
    __branch__ = (
        subprocess.run(
            ["git", "branch", "--show-current"],
            capture_output=True,
            cwd=here,
            check=True,
        )
        .stdout.strip()
        .decode()
    )
except:
    __branch__ = ""

if __branch__:
    __version__ += "+" + __branch__



"""Class representing single property, see YEP-111."""


Getter = object()


class Property:
    def __init__(self, client, prop):
        self._client = client
        self._property = prop

        for k in ("units", "options", "limits"):
            if prop[f"{k}_getter"] is not None:
                setattr(self, k, getattr(self._client, self._property[f"{k}_getter"]))

    def __call__(self, val=Getter):
        if val is Getter:
            return getattr(self._client, self._property["getter"])()
        if not self._property["setter"]:
            raise TypeError("Property is not settable")
        return getattr(self._client, self._property["setter"])(val)

    @property
    def control_kind(self):
        return self._property["control_kind"]

    @property
    def record_kind(self):
        return self._property["record_kind"]

    @property
    def dynamic(self):
        return self._property["dynamic"]

    @property
    def type(self):
        return self._property["type"]



__all__ = ["Socket"]


import socket
import io
import struct

import fastavro  # type: ignore

from ._schema import handshake_request, handshake_response


BUFFSIZE = 4096


class Socket:
    def __init__(self, host, port):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.settimeout(None)
        self._socket.connect((host, port))
        self._named_types = {}

    def _read(self, response_schema):
        response_schema = fastavro.parse_schema(
            response_schema, expand=True, named_schemas=self._named_types
        )
        # Needed twice for nested types, likely should be fixed upstream
        response_schema = fastavro.parse_schema(
            response_schema, expand=True, named_schemas=self._named_types
        )
        buf = io.BytesIO()
        remaining = 0
        while True:
            try:
                buf.seek(0)
                obj = fastavro.schemaless_reader(buf, response_schema)
                return obj
            except Exception:
                buf.seek(0)
            if not remaining:
                remaining = struct.unpack_from(">L", self._socket.recv(4))[0]

            buf.seek(0, 2)
            num_read = buf.write(self._socket.recv(min(remaining, BUFFSIZE)))
            remaining -= num_read

    def _write(self, bytesio):
        bytesio.seek(0)
        out = bytesio.read()
        out = struct.pack(">L", len(out)) + out
        self._socket.sendall(out)

    def handshake(
        self, client_hash=b" " * 16, client_protocol=None, server_hash=b" " * 16
    ):
        # send request
        request = io.BytesIO()
        record = {
            "clientHash": client_hash,
            "clientProtocol": client_protocol,
            "serverHash": server_hash,
            "meta": {},
        }
        fastavro.schemaless_writer(request, handshake_request, record)
        self._write(request)
        self._write_metadata()
        self._write_method_name("")
        # read response
        response = self._read(handshake_response)
        self._read({"type": "map", "values": "bytes"})
        self._read("boolean")
        self._read("null")
        if response["match"] == "NONE":
            self.handshake(
                response["serverHash"],
                response["serverProtocol"],
                response["serverHash"],
            )
        return response["serverProtocol"]

    def message(self, method_name, method_schema, *args, **kwargs):
        self._validate_parameters(method_schema.get("request", []), *args, **kwargs)
        self._write_metadata()
        self._write_method_name(method_name)
        self._write_parameters(method_schema.get("request", []), *args, **kwargs)
        self._write_terminator()
        # read metadata
        _ = self._read({"type": "map", "values": "bytes"})
        # read error
        error = self._read("boolean")
        if error:
            raise Exception(self._read(["string"]))
        # read response
        response = self._read(method_schema.get("response", "null"))
        return response

    def _write_metadata(self, meta=None):
        if meta is None:
            meta = {}
        # write metadata
        out = io.BytesIO()
        fastavro.schemaless_writer(
            out, {"type": "map", "values": "bytes"}, meta
        )  # empty mapping
        self._write(out)

    def _write_method_name(self, method_name):
        # write method_name
        out = io.BytesIO()
        fastavro.schemaless_writer(out, "string", method_name)
        self._write(out)

    def _validate_parameters(self, method_request_schema, *args, **kwargs):
        # write parameters
        args = list(args)
        for parameter in method_request_schema:
            if parameter["name"] in kwargs:
                data = kwargs[parameter["name"]]
            elif args:
                data = args.pop(0)
            out = io.BytesIO()
            schema = fastavro.parse_schema(
                parameter["type"], expand=True, named_schemas=self._named_types
            )
            # Needed twice for nested types... Should likely be fixed upstream
            schema = fastavro.parse_schema(
                schema, expand=True, named_schemas=self._named_types
            )
            fastavro.schemaless_writer(out, schema, data)

    def _write_parameters(self, method_request_schema, *args, **kwargs):
        # write parameters
        args = list(args)
        for parameter in method_request_schema:
            if parameter["name"] in kwargs:
                data = kwargs[parameter["name"]]
            elif args:
                data = args.pop(0)
            out = io.BytesIO()
            schema = fastavro.parse_schema(
                parameter["type"], expand=True, named_schemas=self._named_types
            )
            # Needed twice for nested types... Should likely be fixed upstream
            schema = fastavro.parse_schema(
                schema, expand=True, named_schemas=self._named_types
            )
            fastavro.schemaless_writer(out, schema, data)
            self._write(out)

    def _write_terminator(self):
        # write terminate (zero length buffer)
        self._write(io.BytesIO())
